import os

from datetime import date

from pybman import data
from pybman import utils


class LocalData:

    def __init__(self, base_dir='./data/', ous_dir='ous', ctx_dir='ctx', pers_dir='pers', create=False):

        self.data_dir = os.path.realpath(base_dir)
        # self.data_file = os.path.join(self.data_dir, 'paths.txt')
        self.data_exists = True

        self.ou_dir = os.path.join(self.data_dir, ous_dir)
        self.ctx_dir = os.path.join(self.data_dir, ctx_dir)
        self.pers_dir = os.path.join(self.data_dir, pers_dir)

        self.ou_exists = True
        self.ctx_exists = True
        self.pers_exists = True

        self.data_paths = []

        if not os.path.exists(self.data_dir):
            self.data_exists = False
            self.ou_exists = False
            self.ctx_exists = False
            self.pers_exists = False
            if create:
                os.mkdir(self.data_dir)
                os.mkdir(self.ou_dir)
                os.mkdir(self.ctx_dir)
                os.mkdir(self.pers_dir)

        if not os.path.exists(self.ou_dir) and self.ou_exists:
            self.ou_exists = False
            if create:
                os.mkdir(self.ou_dir)

        if not os.path.exists(self.ctx_dir) and self.ctx_exists:
            self.ctx_exists = False
            if create:
                os.mkdir(self.ctx_dir)

        if not os.path.exists(self.pers_dir) and self.pers_exists:
            self.pers_exists = False
            if create:
                os.mkdir(self.pers_dir)

        if self.data_exists:
            for root, dirs, files in os.walk(self.data_dir):
                for name in files:
                    if name.endswith(".txt"):
                        self.data_paths.append(os.path.join(root, name))
                    elif name.endswith(".json"):
                        self.data_paths.append(os.path.join(root, name))
                    elif name.endswith(".csv"):
                        self.data_paths.append(os.path.join(root, name))
                    else:
                        continue
            if self.data_paths:
                self.data_paths.sort()
                print('local pulication data:')
                for p in self.data_paths[:25]:
                    print(p)
                if len(self.data_paths) > 25:
                    print(". . .")

    # find path by given pattern
    def find_data_path(self, pattern):
        found_paths = []
        if self.data_paths:
            for p in self.data_paths:
                if pattern in p:
                    found_paths.append(p)
            if not found_paths:
                print("could not find path containing", pattern)
            return found_paths
        else:
            print('no local data!')
            return found_paths

    # get local data
    def get_data(self, pattern):
        data_sets = []
        paths = self.find_data_path(pattern)
        if paths:
            for path in paths:
                json_data = utils.read_json(path)
                data_idx = path.split("/")[-1].split(".")[0]
                data_sets.append(data.DataSet(data_idx, data=json_data))
        else:
            print("could not find local data!")
        return data_sets

    # write local data file
    def store_data(self, idx, dict_data):
        print("store local data of", idx)
        # self.change_data_path(idx)
        path = self.generate_data_path(idx)
        utils.write_json(path, dict_data)

    # get data path for given id
    def generate_data_path(self, data_id):
        return self.data_dir + data_id + "--" + date.today().isoformat() + ".json"

    # def change_data_path(self, idx):
    #    """
    #    update list of paths
    #    """
    #    pos = -1
    #    for i, path in enumerate(self.data_paths):
    #        if idx in path:
    #            pos = i
    #            break
    #    new = self.generate_data_path(idx)
    #    if pos < 0:
    #        self.data_paths.append(new)
    #        utils.write_list(self.data_file, self.data_paths)
    #        print("local data file added", new)
    #    else:
    #        old = self.data_paths[pos]
    #        # self.clean_local_data(old)
    #        self.data_paths[pos] = new
    #        utils.write_list(self.data_file, self.data_paths)
    #        print("local data file updated", new)

    # remove old local data
    # def clean_local_data(self, idx):
    #    """
    #    clean up local data
    #    """
    #    path = self.find_data_path(idx)
    #    if path:
    #        print("removing file", path)
    #        os.remove(path)
    #    else:
    #        print("failed to remove file from entity", idx)

#    def store_titles_local(self, idx, data):
#        pass



# from pybman import local
# from pybman import export


class DataSet:

    def __init__(self, data_id, data=None, raw=None):

        self.idx = data_id

        if raw:
            self.collection = {"numberOfRecords": len(raw), "records": raw}
            self.num = len(raw)
            self.records = raw
            self.persons = self.get_cone_persons()

        elif data:
            self.collection = data
            self.num = len(data['records'])
            self.records = data['records']
            self.persons = self.get_cone_persons()

        else:
            # print("please pass (raw) data for initialization!")
            self.collection = {}
            self.num = 0
            self.records = []
            self.persons = {}
            # self.creators = []

    # get creators of items
    def get_creators(self):
        creators = []
        for record in self.records:
            if 'creators' in record['data']['metadata']:
                creators_list = record['data']['metadata']['creators']
                for creator in creators_list:
                    creators.append(creator)
        return creators

    def get_creators_from_records(self):
        creators = []
        for record in self.records:
            if 'creators' in record['data']['metadata']:
                creators.append(record['data']['metadata']['creators'])
        return creators

    def get_creators_data(self):
        creators = {}
        for record in self.records:
            if 'creators' in record['data']['metadata']:
                creators_list = record['data']['metadata']['creators']
                found = False
                for creator in creators_list:
                    if 'person' in creator:
                        found = True
                        if 'identifier' in creator['person']:
                            idx = creator['person']['identifier']['id'].split("/")[-1]
                            if idx in creators:
                                creators[idx].append(record)
                            else:
                                creators[idx] = [record]
                        else:
                            continue
                            # if 'givenName' in creator['person'] and 'familyName' in creator['person']:
                            #    pers_name = creator['person']['givenName'] + " " + creator['person']['familyName']
                            #    # print("no identifier found for", pers_name)
                            # elif 'givenName' not in person and 'familyName' in person:
                            #    # pers_name = creator['person']['familyName']
                            #    # print("no identifier found for", pers_name,"of",record['data']['objectId'])
                            # else:
                            #    # print("no identifier found for creator of", record['data']['objectId'])
                if not found:
                    print("no person found for", record['data']['objectId'])
        return creators

    def get_cone_persons(self):
        persons = {}
        for creator in self.get_creators():
            if creator['type'] == 'PERSON':
                if 'identifier' in creator['person']:
                    if creator['person']['identifier']['type'] == 'CONE':
                        if creator['person']['identifier']['id'] not in persons:
                            cone_id = creator['person']['identifier']['id'].split("/")[-1]
                            if 'givenName' in creator['person']:
                                name = creator['person']['givenName']
                                if 'familyName' in creator['person']:
                                    name += ' ' + creator['person']['familyName']
                            else:
                                if 'familyName' in creator['person']:
                                    name = creator['person']['familyName']
                                else:
                                    print('no name found for', cone_id)
                                    name = ''
                            persons[cone_id] = name
        return persons


#    def init_pers_data(self):
#        pers_data = []
#        for p in self.persons:
#            p_data = export.get_pers(p)
#            p_idx = p.split('/')[-1]
#            pers_data.append(DataSet(p_idx,data=p_data))
#        return pers_data

    def get_organizations(self):
        organizations = {}
        for record in self.records:
            if 'creators' in record['data']['metadata']:
                creators_list = record['data']['metadata']['creators']
                for creator in creators_list:
                    if creator['type'] == 'PERSON':
                        person = creator['person']
                        if 'organizations' in person:
                            for organization in person['organizations']:
                                if organization['name'] in organizations:
                                    organizations[organization['name']].append(record['data']['objectId'])
                                else:
                                    organizations[organization['name']] = [record['data']['objectId']]
                        else:
                            print("no organization found for", person)
                    elif creator['type'] == 'ORGANIZATION':
                        organization = creator['organization']
                        if organization['name'] in organizations:
                            organizations[organization['name']].append(record['data']['objectId'])
                        else:
                            organizations[organization['name']] = [record['data']['objectId']]
                    else:
                        print("unknown creator type", creator['type'])
            else:
                print(record['data']['objectId'], "has no creator!")
        return organizations

    # get titles of items
    def get_titles(self):
        titles = {}
        for record in self.records:
            title = record['data']['metadata']['title']
            if title in titles:
                titles[title].append(record['data']['objectId'])
            else:
                titles[title] = [record['data']['objectId']]
            # titles.append(record['data']['metadata']['title'])
        return titles

    # get titles from source of items
    def get_titles_from_source(self):
        source_titles = {}
        for record in self.records:
            if 'sources' in record['data']['metadata']:
                for source in record['data']['metadata']['sources']:
                    if source['title'] in source_titles:
                        source_titles[source['title']].append(record['data']['objectId'])
                    else:
                        source_titles[source['title']] = [record['data']['objectId']]
                    # source_titles.append(source['title'])
        return source_titles

    # get list of genres from collection
    def get_genres(self):
        genres = {}
        for record in self.records:
            if record['data']['metadata']['genre'] in genres:
                genres[record['data']['metadata']['genre']].append(record['data']['objectId'])
            else:
                genres[record['data']['metadata']['genre']] = [record['data']['objectId']]
        return genres

    # get list of genres from collection
    def get_genre_data(self):
        genres = {}
        for record in self.records:
            if record['data']['metadata']['genre'] in genres:
                genres[record['data']['metadata']['genre']].append(record)
            else:
                genres[record['data']['metadata']['genre']] = [record]
        return genres

    # get publication places of items
    def get_places(self):
        places = {}
        for record in self.records:
            if 'publishingInfo' in record['data']['metadata']:
                if 'place' in record['data']['metadata']['publishingInfo']:
                    place = record['data']['metadata']['publishingInfo']['place']
                    if place in places:
                        places[place].append(record['data']['objectId'])
                    else:
                        places[place] = [record['data']['objectId']]
            if 'sources' in record['data']['metadata']:
                for source in record['data']['metadata']['sources']:
                    if 'publishingInfo' in source:
                        if 'place' in source['publishingInfo']:
                            place = source['publishingInfo']['place']
                            if place in places:
                                places[place].append(record['data']['objectId'])
                            else:
                                places[place] = [record['data']['objectId']]
        return places

    # get publishers of items
    def get_publishers(self):
        publishers = {}
        for record in self.records:
            if 'publishingInfo' in record['data']['metadata']:
                if 'publisher' in record['data']['metadata']['publishingInfo']:
                    publisher = record['data']['metadata']['publishingInfo']['publisher']
                    if publisher in publishers:
                        publishers[publisher].append(record['data']['objectId'])
                    else:
                        publishers[publisher] = [record['data']['objectId']]
            if 'sources' in record['data']['metadata']:
                for source in record['data']['metadata']['sources']:
                    if 'publishingInfo' in source:
                        if 'publisher' in source['publishingInfo']:
                            publisher = source['publishingInfo']['publisher']
                            if publisher in publishers:
                                publishers[publisher].append(record['data']['objectId'])
                            else:
                                publishers[publisher] = [record['data']['objectId']]
        return publishers

    def get_journals(self):
        journals = {}
        items = self.get_items_with_source_genre("JOURNAL")
        for k in items:
            if items[k]['title'] in journals:
                journals[items[k]['title']].append(k)
            else:
                journals[items[k]['title']] = [k]
        return journals

    def get_journals_data(self):
        journals = {}
        items = self.get_items_with_source_genre("JOURNAL")
        for k in items:
            if items[k]['title'] in journals:
                journals[items[k]['title']].append(self.get_item(k))
            else:
                journals[items[k]['title']] = [self.get_item(k)]
        return journals

    def get_series(self):
        series = {}
        items = self.get_items_with_source_genre("SERIES")
        for k in items:
            if items[k]['title'] in series:
                series[items[k]['title']].append(k)
            else:
                series[items[k]['title']] = [k]
        return series

    def get_years(self):
        years = {}
        for record in self.records:
            if 'datePublishedInPrint' in record['data']['metadata']:
                year = record['data']['metadata']['datePublishedInPrint'].split("-")[0]
                if year in years:
                    years[year].append(record['data']['objectId'])
                else:
                    years[year] = [record['data']['objectId']]
            elif 'datePublishedOnline' in record['data']['metadata']:
                year = record['data']['metadata']['datePublishedOnline'].split("-")[0]
                if year in years:
                    years[year].append(record['data']['objectId'])
                else:
                    years[year] = [record['data']['objectId']]
            else:
                print("no publication date found for", record['data']['objectId'])
        return years

    def get_years_data(self):
        years = {}
        for record in self.records:
            if 'datePublishedInPrint' in record['data']['metadata']:
                year = record['data']['metadata']['datePublishedInPrint'].split("-")[0]
                if year in years:
                    years[year].append(record)
                else:
                    years[year] = [record]
            elif 'datePublishedOnline' in record['data']['metadata']:
                year = record['data']['metadata']['datePublishedOnline'].split("-")[0]
                if year in years:
                    years[year].append(record)
                else:
                    years[year] = [record]
            else:
                print("no publication date found for", record['data']['objectId'])
        return years

    def get_languages_data(self):
        languages = {}
        for record in self.records:
            if 'languages' in record['data']['metadata']:
                lang = record['data']['metadata']['languages']
                if len(lang) == 1:
                    if lang[0] in languages:
                        languages[lang[0]].append(record)
                    else:
                        languages[lang[0]] = [record]
                elif len(lang) > 1:
                    # print(record['data']['objectId'], "has more than one language!")
                    for l in lang:
                        if l in languages:
                            languages[l].append(record)
                        else:
                            languages[l] = [record]
                else:
                    print(record['data']['objectId'], "has no language!")
        return languages

    # get identifiers from sources of items
    def get_sources_identifiers(self):
        source_idx = {}
        for record in self.records:
            if 'sources' in record['data']['metadata']:
                for source in record['data']['metadata']['sources']:
                    if 'identifiers' in source:
                        for identifier in source['identifiers']:
                            identifier_type = identifier['type']
                            if identifier_type in source_idx:
                                source_idx[identifier_type].append(record['data']['objectId'])
                            else:
                                source_idx[identifier_type] = [record['data']['objectId']]
        return source_idx

    def get_sources_titles(self, genre=None):
        source_titles = {}
        for record in self.records:
            if 'sources' in record['data']['metadata']:
                for source in record['data']['metadata']['sources']:
                    title = source['title']
                    if title in source_titles:
                        source_titles[title].append(record['data']['objectId'])
                    else:
                        source_titles[title] = [record['data']['objectId']]
        return source_titles

    # get items with state 'submitted'
    def get_items_submitted(self):
        submitted = {}
        for record in self.records:
            if record['data']['publicState'] == 'SUBMITTED':
                if record['data']['versionState'] == 'SUBMITTED':
                    submitted[record['persistenceId']] = record['data']
        return submitted

    # get items with state 'submitted'
    def get_items_released(self):
        released = []
        for record in self.records:
            if record['data']['publicState'] == 'RELEASED':
                if record['data']['versionState'] == 'RELEASED':
                    released.append(record)
        return released

    # get item from collection by given id
    def get_item(self, item_id):
        item = {}
        for record in self.records:
            if item_id == record['data']['objectId']:
                item = record
                break
        return item

    # get items from collection with given genre
    def get_items_with_genre(self, genre):
        records = {}
        for record in self.records:
            if genre == record['data']['metadata']['genre']:
                records[record['data']['objectId']] = record
        return records

    def get_items_with_publication_year(self, year):
        # dateAccepted
        # dateCreated
        # dateModified
        # datePublishedInPrint
        # datePublishedOnline
        # dateSubmitted
        pass

    # get items from collection with given source genre
    def get_items_with_source_genre(self, source_genre):
        records = {}
        for record in self.records:
            if 'sources' in record['data']['metadata']:
                for source in record['data']['metadata']['sources']:
                    if source_genre == source['genre']:
                        records[record['data']['objectId']] = source
        return records

    # get items from collection with external url
    def get_items_with_external_url(self):
        records = {}
        for record in self.records:
            if 'files' in record['data']:
                for f in record['data']['files']:
                    if f['storage'] == 'EXTERNAL_URL':
                        records[record['data']['objectId']] = record
        return records

    # get items from collection with uri as identifier
    def get_items_with_identifier_uri(self):
        records = {}
        for record in self.records:
            if 'identifiers' in record['data']['metadata']:
                identifiers = record['data']['metadata']['identifiers']
                for idx in identifiers:
                    if idx['type'] == 'URI':
                        records[record['data']['objectId']] = record
        return records

    def save_titles(self):
        # titles = self.get_titles()
        pass



import re
import json
import requests
# import urllib3
import pkg_resources

from urllib.parse import urlencode


def clean_string(string):
    string = string.strip()
    string = string.replace("\n", " ")
    string = string.replace("\r", " ")
    string = re.sub(' +', ' ', string)
    return string

# write given list (results) to file at path
def write_list(path, results):
    print("write list to file", path)
    if type(results[0]) == str:
        with open(path, "w+", encoding="utf8") as f:
            f.write("\n".join(results))
    if type(results[0]) == list:
        with open(path, "w+", encoding="utf8") as f:
            for res in results:
                f.write('"' + '"\n"'.join(res) + '"\n')


def write_csv(path, results):
    print("write csv to file", path)
    with open(path, "w+", encoding="utf8") as f:
        for row in results:
            f.write('"' + '","'.join(row) + '"\n')


# read plain text file
def read_plain_clean(path):
    print("read plain text file", path)
    lines = []
    with open(path, 'r', encoding='utf-8') as f:
        for line in f.readlines():
            lines.append(line.strip('\n'))
    return lines


# read json file at path
def read_json(path):
    print("read file", path)
    result = {}
    with open(path, 'r', encoding='utf-8') as f:
        result = json.load(f)
    return result


# write given data to json file at path
def write_json(path, data):
    print("write to", path)
    with open(path, 'w', encoding="utf-8") as f:
        s = json.dumps(data, indent=2)
        f.write(s)
    return path


# send get request to fetch data
def get_request(url, params=None, headers=None, json_response=True):
    if params:
        params = urlencode(params)
        url = url + params
    if headers:
        response = requests.get(url, headers=headers)
    else:
        response = requests.get(url)
    if response.status_code == 200:
        if json_response:
            return response.json()
        else:
            return response
    else:
        # print("something went wrong while requesting data!")
        print("got status code", response.status_code, "for url:\n", url)
        return {}


# send a post request with data
def post_request(url, params=None, headers=None, data=None, json_res=True):
    if params:
        params = urlencode(params)
        url = url + params
    if type(data) == dict:
        payload = json.dumps(data)
    else:
        payload = data
    response = requests.post(url, headers=headers, data=payload)
    if response.status_code == 200:
        if json_res:
            return response.json()
        else:
            return response
    else:
        # print("something went wrong while requesting data!")
        print("got status code", response.status_code, "for url:\n", url)
        return {}


# send a put request with data
def put_request(url, header, data):
    payload = json.dumps(data)
    response = requests.put(url, headers=header, data=payload)
    if response.status_code == 200:
        return response.json()
    else:
        print("something went wrong while requesting data!")
        print("got status code", response.status_code, "for url:\n", url)
        return {}


user_agent = {'User-Agent': 'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:64.0) Gecko/20100101 Firefox/64.0'}


def check_url(url):
    try:
        response = requests.head(url)
    except Exception as e:
        print(str(e))
        print("... tried to access:", url)
        return ''
    if response.status_code == 200:
        # print("HTTP 200", url)
        return url
    elif response.status_code == 303:
        # print("HTTP 302", url)
        return response.headers['Location']
    elif response.status_code == 302:
        # print("HTTP 302", url)
        return response.headers['Location']
    elif response.status_code == 301:
        # print("HTTP 301", url)
        return response.headers['Location']  # check if 301 has new location
    elif response.status_code == 403:
        print("HTTP", response.status_code, url)
        print("could not check")
        return url
    elif response.status_code == 404:
        return ''
    else:
        print("HTTP", response.status_code, url)
        print("... unhandled status code!")
        return url


def url_exists(url):
    try:
        response = get_request(url, json_response=False, headers=user_agent)
    except Exception as e:
        print(str(e))
        print("while accessing", url)
        return False
    if not response:
        return False
    else:
        return True


def url_exists2(url):
    try:
        response = requests.head(url)
    except Exception as e:
        print(str(e))
        print("... tried to access:", url)
        return ''
    if response.status_code == 404:
        return False
    else:
        return True


# resolve path of package files
def resolve_path(path):
    return pkg_resources.resource_filename('pybman', path)



from pybman import data
# from pybman import local
from pybman import query
from pybman import rest
from pybman import inspector
# from pybman import utils


class Client:

    def __init__(self, secret=''):

        # load rest api controllers
        self.item_rest = rest.ItemRestController(secret)
        self.ou_rest = rest.OrgUnitRestController()
        self.ctx_rest = rest.ContextRestController()
        self.feed_rest = rest.FeedRestController()

        # load cone controllers
        self.journals_cone = rest.JournalConeController()
        self.persons_cone = rest.PersonConeController()
        self.lang_cone = rest.LanguageConeController()

        # load classes providing search queries
        self.pers_query = query.PersQuery()
        self.ou_query = query.OrgUnitQuery()
        self.ctx_query = query.ContextQuery()
        self.lang_query = query.LangQuery()
        self.jour_query = query.JournalQuery()
        self.all_query = query.AllQuery()

        # inspector class
        self.inspector = None

    def get_data(self, ctx_id=None, ou_id=None, pers_id=None, lang_id=None, jour_name=None, misc_query=None):
        if ctx_id:
            ctx_query = self.ctx_query.get_item_query(ctx_id)
            self.item_rest.search_items(query=ctx_query)
            return data.DataSet(ctx_id, raw=self.item_rest.records)
        elif ou_id:
            ou_query = self.ou_query.get_item_query(ou_id)
            self.item_rest.search_items(query=ou_query)
            return data.DataSet(ou_id, raw=self.item_rest.records)
        elif pers_id:
            pers_query = self.pers_query.get_item_query(pers_id)
            self.item_rest.search_items(query=pers_query)
            return data.DataSet(pers_id, raw=self.item_rest.records)
        elif lang_id:
            lang_query = self.lang_query.get_item_query(lang_id)
            self.item_rest.search_items(query=lang_query)
            return data.DataSet(lang_id, raw=self.item_rest.records)
        elif jour_name:
            jour_query = self.jour_query.get_item_query(jour_name)
            self.item_rest.search_items(query=jour_query)
            return data.DataSet(jour_name, raw=self.item_rest.records)
        elif misc_query:
            self.item_rest.search_items(query=misc_query)
            return data.DataSet("query_data", raw=self.item_rest.records)
        else:
            print("please specify data to retrieve!")

    def update_data(self, idx, item_data, comment):
        updated_data = self.item_rest.update_item(idx, item_data)
        if updated_data:
            response = self.item_rest.release_item(idx, updated_data, comment)
            if response:
                print("successfully updated data with id", response['objectId'])
                return response
            else:
                print("something went wrong while releasing item with id", idx)
                return None
        else:
            print("something went wrong while updating item with id", idx)
            return None

    def clean_init(self, data_set):
        self.inspector = inspector.Inspector(data_set.records)

    def clean_titles(self, data_set=None):

        print("start cleaning title data!")

        # create new inspector instance
        if data_set:
            self.clean_init(data_set)
        else:
            print("failed to initialize inspector!")
            print("please pass data to be cleaned.")
            return 0

        # cleaning of publication titles
        clean_data = self.inspector.check_publication_titles(clean=True)
        total = 0
        if clean_data:
            comment = 'auto-update: publication title stripped'
            for k in clean_data:
                updated = self.update_data(k, clean_data[k]['data'], comment)
                if updated:
                    total += 1
            print("updated", total, "publication titles!")
        else:
            print("publication title data is already clean!")
            print("nothing to do...")

        return total

    def change_genre(self, new_genre, old_genre, data_set=None):

        print("start changing genre of items")

        # create new inspector instance
        if data_set:
            self.clean_init(data_set)
        else:
            print("failed to initialize inspector!")
            print("please pass data to be cleaned.")
            return 0

        clean_data = self.inspector.change_genre(new_genre, old_genre)
        total = 0
        if clean_data:
            comment = 'auto-update: change genre of item from ' + old_genre + " to " + new_genre
            for k in clean_data:
                updated = self.update_data(k, clean_data[k]['data'], comment)
                if updated:
                    total += 1
            print("updated genre of", total, "items!")
        else:
            print("genre is correctly chosen already!")
            print("nothing to do...")
            return 0

    def clean_source_titles(self, data_set=None):

        print("start cleaning source title data!")

        # create new inspector instance
        if data_set:
            self.clean_init(data_set)
        else:
            print("failed to initialize inspector!")
            print("please pass data to be cleaned.")
            return 0

        # cleaning of source title data
        clean_data = self.inspector.check_source_titles(clean=True)
        total = 0
        if clean_data:
            comment = 'auto-update: source title stripped'
            for k in clean_data:
                updated = self.update_data(k, clean_data[k]['data'], comment)
                if updated:
                    total += 1
            print("updated", total, "source titles!")
        else:
            print("source title data is already clean!")
            print("nothing to do...")

        return total

    def clean_publishers(self, data_set=None):

        print("start cleaning publisher data!")

        # create new inspector instance
        if data_set:
            self.clean_init(data_set)
        else:
            print("failed to initialize inspector!")
            print("please pass data to be cleaned.")
            return 0

        # cleaning of publisher data
        clean_data = self.inspector.check_publishers(clean=True)
        total = 0
        if clean_data:
            comment = 'auto-update: remove control characters from publisher value'
            for k in clean_data:
                updated = self.update_data(k, clean_data[k]['data'], comment)
                if updated:
                    total += 1
        else:
            print("publisher data is already clean!")
            print("nothing to do...")

        print("updated", total, "items!")
        return total

    def clean_publishing_places(self, data_set=None):

        print("start cleaning publishing place data!")

        # create new inspector instance
        if data_set:
            self.clean_init(data_set)
        else:
            print("failed to initialize inspector!")
            print("please pass data to be cleaned.")
            return 0

        # start cleaning of publishing place data
        clean_data = self.inspector.check_publishing_places(clean=True)
        total = 0
        if clean_data:
            comment = 'auto-update: remove control characters from publishing place value'
            for k in clean_data:
                updated = self.update_data(k, clean_data[k]['data'], comment)
                if updated:
                    total += 1
        else:
            print("publishing place data is already clean!")
            print("nothing to do...")

        print("updated", total, "items!")
        return total



from copy import deepcopy
from pybman import utils


class Query:

    def __init__(self):
        pass


class ConeQuery:

    def __init__(self):
        pass

class AllQuery:

    def __init__(self):
        self.files_query_fp = utils.resolve_path('static/elastic/item-all-files-released.json')
        self.locators_query_fp = utils.resolve_path('static/elastic/item-all-locators-released.json')

        self.files_query = utils.read_json(self.files_query_fp)
        self.locators_query = utils.read_json(self.locators_query_fp)

    def get_files_query(self):
        data = deepcopy(self.files_query)
        return data

    def get_locators_query(self):
        data = deepcopy(self.locators_query)
        return data

class ContextQuery:

    def __init__(self):

        self.item_query_fp = utils.resolve_path('static/elastic/item-ctx.json')
        self.item_released_query_fp = utils.resolve_path('static/elastic/item-ctx-released.json')

        self.item_query = utils.read_json(self.item_query_fp)
        self.item_released_query = utils.read_json(self.item_released_query_fp)

    def get_item_query(self, ctx_id):
        data = deepcopy(self.item_query)
        data['query']['term']['context.objectId']['value'] = ctx_id
        return data

    def get_released_item_query(self, ctx_id):
        data = deepcopy(self.item_released_query)
        data['query']['bool']['must'][2]['term']['context.objectId']['value'] = ctx_id
        return data


class OrgUnitQuery:

    def __init__(self):

        self.item_query_fp = utils.resolve_path('static/elastic/item-ou.json')
        self.item_released_query_fp = utils.resolve_path('static/elastic/item-ou-released.json')

        self.item_query = utils.read_json(self.item_query_fp)
        self.item_released_query = utils.read_json(self.item_released_query_fp)

    def get_item_query(self, ou_id):
        data = deepcopy(self.item_query)
        term = data['query']['bool']['should'][0]['term']
        term['metadata.creators.person.organizations.identifierPath']['value'] = ou_id
        term = data['query']['bool']['should'][1]['term']
        term['metadata.creators.organization.identifierPath']['value'] = ou_id
        return data

    def get_item_released_query(self, ou_id):
        data = deepcopy(self.item_released_query)
        term = data['query']['bool']['must'][2]['bool']['should'][0]['term']
        term['metadata.creators.person.organizations.identifierPath']['value'] = ou_id
        term = data['query']['bool']['must'][2]['bool']['should'][1]['term']
        term['metadata.creators.organization.identifierPath']['value'] = ou_id
        return data


class PersQuery:

    def __init__(self):

        self.item_query_fp = utils.resolve_path('static/elastic/item-pers.json')
        self.item_released_query_fp = utils.resolve_path('static/elastic/item-pers-released.json')

        self.item_query = utils.read_json(self.item_query_fp)
        self.item_released_query = utils.read_json(self.item_released_query_fp)

        self.cone_id_format = '/persons/resource/'

    def get_item_query(self, cone_id):
        data = deepcopy(self.item_query)
        data['query']['term']['metadata.creators.person.identifier.id']['value'] = self.cone_id_format + cone_id
        return data

    def get_item_released_query(self, cone_id):
        data = deepcopy(self.item_released_query)
        term = data['query']['bool']['must'][2]['term']
        term['metadata.creators.person.identifier.id']['value'] = self.cone_id_format + cone_id
        return data


class LangQuery:

    def __init__(self):
        self.item_query_fp = utils.resolve_path('static/elastic/item-lang.json')
        self.item_released_query_fp = utils.resolve_path('static/elastic/item-lang-released.json')

        self.item_query = utils.read_json(self.item_query_fp)
        self.item_released_query = utils.read_json(self.item_released_query_fp)

    def get_item_query(self, lang_id):
        data = deepcopy(self.item_query)
        data['query']['term']['metadata.languages']['value'] = lang_id
        return data

    def get_released_item_query(self, lang_id):
        data = deepcopy(self.item_released_query)
        data['query']['bool']['must'][2]['term']['metadata.languages']['value'] = lang_id
        return data

class JournalQuery:

    def __init__(self):
        self.item_query_fp = utils.resolve_path('static/elastic/item-jour.json')
        self.item_released_query_fp = utils.resolve_path('static/elastic/item-jour-released.json')

        self.item_query = utils.read_json(self.item_query_fp)
        self.item_released_query = utils.read_json(self.item_released_query_fp)

    def get_item_query(self, jour_name):
        data = deepcopy(self.item_query)
        data['query']['bool']['should'][0]['match_phrase']['metadata.sources.title']['query'] = jour_name
        data['query']['bool']['should'][1]['match_phrase']['metadata.sources.alternativeTitles.value']['query'] = jour_name
        return data

    def get_released_item_query(self, jour_name):
        data = deepcopy(self.item_released_query)
        data['query']['bool']['must'][2]['bool']['should'][0]['match_phrase']['metadata.sources.title']['query'] = jour_name
        data['query']['bool']['must'][2]['bool']['should'][1]['match_phrase']['metadata.sources.alternativeTitles.value']['query'] = jour_name
        return data



from pybman import utils
# from pybman import rest


class Inspector:

    def __init__(self, records):
        self.records = records

    def check_publication_titles(self, clean=False):
        updates = {}
        for record in self.records:
            publication_title = record['data']['metadata']['title']
            # if publication_title != publication_title.strip():
            #    item_id = record['data']['objectId']
            #    if clean:
            #        record['data']['metadata']['title'] = publication_title.strip()
            #    updates[item_id] = record
            if publication_title != utils.clean_string(publication_title):
                item_id = record['data']['objectId']
                if clean:
                    # print("dirty:", publication_title)
                    record['data']['metadata']['title'] = utils.clean_string(publication_title)
                    # print(record['data']['metadata']['title'])
                updates[item_id] = record
        return updates

    def check_source_titles(self, clean=False):
        updates = {}
        for record in self.records:
            if 'sources' in record['data']['metadata']:
                for source in record['data']['metadata']['sources']:
                    if source['title'] != utils.clean_string(source['title']):
                        item_id = record['data']['objectId']
                        if clean:
                            # print("dirty:",source['title'])
                            source['title'] = utils.clean_string(source['title'])
                            # print("clean:", source['title'])
                        updates[item_id] = record
        return updates

    def check_publishers(self, clean=False):
        updates = {}
        for record in self.records:
            if 'publishingInfo' in record['data']['metadata']:
                if 'publisher' in record['data']['metadata']['publishingInfo']:
                    publisher = record['data']['metadata']['publishingInfo']['publisher']
                    if publisher != utils.clean_string(publisher):
                        item_id = record['data']['objectId']
                        if clean:
                            record['data']['metadata']['publishingInfo']['publisher'] = utils.clean_string(publisher)
                        updates[item_id] = record
            if 'sources' in record['data']['metadata']:
                for source in record['data']['metadata']['sources']:
                    if 'publishingInfo' in source:
                        if 'publisher' in source['publishingInfo']:
                            publisher = source['publishingInfo']['publisher']
                            if publisher != utils.clean_string(publisher):
                                item_id = record['data']['objectId']
                                if clean:
                                    source['publishingInfo']['publisher'] = utils.clean_string(publisher)
                                updates[item_id] = record
        return updates

    def check_publishing_places(self, clean=False):
        updates = {}
        for record in self.records:
            if 'publishingInfo' in record['data']['metadata']:
                if 'place' in record['data']['metadata']['publishingInfo']:
                    place = record['data']['metadata']['publishingInfo']['place']
                    if place != utils.clean_string(place):
                        item_id = record['data']['objectId']
                        if clean:
                            record['data']['metadata']['publishingInfo']['place'] = utils.clean_string(place)
                        updates[item_id] = record
            if 'sources' in record['data']['metadata']:
                for source in record['data']['metadata']['sources']:
                    if 'publishingInfo' in source:
                        if 'place' in source['publishingInfo']:
                            place = source['publishingInfo']['place']
                            if place != utils.clean_string(place):
                                item_id = record['data']['objectId']
                                if clean:
                                    source['publishingInfo']['place'] = utils.clean_string(place)
                                updates[item_id] = record
        return updates

    # @staticmethod
    # def clean_publication_titles(records):
    #    for k in records:
    #        publication = records[k]
    #        dirty_title = publication['data']['metadata']['title']
    #        publication['data']['metadata']['title'] = dirty_title.strip()
    #    return records

    def check_publication_date(self):
        # dateAccepted
        # dateCreated
        # dateModified
        # datePublishedInPrint
        # datePublishedOnline
        # dateSubmitted
        pass

    def check_publication_uri(self):
        # was ist wenn zwei uri's auftauchen? eine geht, eine nicht, beide gehen?
        updates = {}
        for record in self.records:
            if 'identifiers' in record['data']['metadata']:
                identifiers = record['data']['metadata']['identifiers']
                for idx in identifiers:
                    if idx['type'] == 'URI':
                        url = idx['id']
                        if not utils.url_exists(url):
                            item_id = record['data']['objectId']
                            updates[item_id] = url
        return updates

    def check_publication_url(self):
        updates = {}
        for record in self.records:
            if 'files' in record['data']:
                for f in record['data']['files']:
                    if f['storage'] == 'EXTERNAL_URL':
                        url = f['content']
                        # check = utils.check_url(url)
                        # if url != check:
                        #    item_id = record['data']['objectId']
                        #    updates[item_id] = check
                        if not utils.url_exists(url):
                            item_id = record['data']['objectId']
                            updates[item_id] = url
                        # if 'name' in f:
                        #    url = f['name']
                        #    if not utils.url_exists(url):
                        #        item_id = record['data']['objectId']
                        #        updates[item_id] = url
                        # else:
                        #    print("")
                        #    print("no name given for", record['data']['objectId'])
                        #    print(record['data']['files'])
                        #    print("")
        return updates

    def change_genre(self, new_genre, old_genre):
        updates = {}
        for record in self.records:
            if old_genre == record['data']['metadata']['genre']:
                record['data']['metadata']['genre'] = new_genre
                updates[record['data']['objectId']] = record
            else:
                print("skipping item", record['data']['objectId'])
        return updates

    def change_pers_name(self, old_family_name=None, new_family_name=None, old_given_name=None, new_given_name=None):
        updates = {}
        if old_family_name and new_family_name:
            for record in self.records:
                creators = record['data']['metadata']['creators']
                for creator in creators:
                    if creator['type'] == 'PERSON':
                        if creator['person']['familyName'] == old_family_name:
                            creator['person']['familyName'] = new_family_name
                            updates[record['data']['objectId']] = record
            return updates

        elif old_given_name and new_given_name:
            for record in self.records:
                creators = record['data']['metadata']['creators']
                for creator in creators:
                    if creator['type'] == 'PERSON':
                        if creator['person']['givenName'] == old_given_name:
                            creator['person']['givenName'] = new_given_name
                            updates[record['data']['objectId']] = record
            return updates

        else:
            print("please pass either new and old family name or given name!")
            return updates



import os
import setuptools

ROOT = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(ROOT, 'README.rst')).read()

with open('requirements.txt') as f:
    required = f.read().splitlines()

setuptools.setup(
    name="pybman",
    version="2019.5.29",
    author="Donatus Herre",
    author_email="pypi@herre.io",
    license="GPL",
    description="Python package (under development) for interacting with MPG.PuRe",
    long_description=README,
    url="https://github.com/herreio/pybman/",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=required,
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
    ],
)



import atexit

from tqdm import tqdm
from copy import deepcopy

from pybman import utils


class BaseController:

    def __init__(self, url='https://pure.mpg.de/'):
        self.records = []

        # base url
        self.url = url
        self.format = {'format': 'json'}
        self.content = {"Content-Type": "application/json"}

        # self.format_query = {'format':'json', 'q':''}
        # self.scroll_query = {"scroll":"true"}


class RestController(BaseController):

    def __init__(self):

        super().__init__()

        # rest interface
        self.rest = self.url + 'rest/'


class LoginRestController(RestController):

    def __init__(self, auth=True, cred=''):

        super().__init__()

        self.auth = auth

        # endpoints
        self.rest_login = self.rest + 'login'
        self.rest_login_who = self.rest_login + '/who'
        self.rest_logout = self.rest + 'logout'

        # user credentials
        self.secret = None

        # user data
        self.token = None
        self.grants = None
        self.roles = None

        # user status
        self.online = False

        self.accept = {'accept': 'application/json'}
        self.header = {'accept': 'application/json', 'Content-Type': 'application/json'}
        self.auth_header = deepcopy(self.accept)

        if auth:
            # read user credentials
            self.secret = utils.read_json(cred)
            # login via REST
            self.login()
            # get permission info
            self.get_user()
            # ensure logout at end of session
            atexit.register(self.logout)

    def login(self):
        if self.secret:
            data = self.secret['user-pass']
            response = utils.post_request(self.rest_login, headers=self.header, data=data, json_res=False)
            self.token = response.headers['Token']
            self.auth_header['Authorization'] = self.token
            self.online = True
        else:
            print("login failed! please provide credentials!")

    def get_user(self):
        if self.online:
            response = utils.get_request(self.rest_login_who, headers=self.auth_header)
            grantlist = response['grantList']
            grants = {}
            for grant in grantlist:
                if grant['role'] in grants:
                    grants[grant['role']].append(grant['objectRef'])
                else:
                    grants[grant['role']] = [grant['objectRef']]
            self.grants = grants
            self.roles = list(self.grants.keys())
            self.roles.sort()
        else:
            print("you need to be logged in to get user info!")

    # func to logout via REST
    def logout(self):
        if self.online:
            response = utils.get_request(self.rest_logout, headers=self.auth_header, json_response=False)
            print(response.text)
            self.online = False
        else:
            print("you need to be logged in to log out!")


class ItemRestController(LoginRestController):

    def __init__(self, secret=''):

        if secret:
            super().__init__(cred=secret)
        else:
            super().__init__(auth=False)

        # items endpoint
        self.rest_items = self.rest + 'items/'
        self.rest_items_release = self.rest_items + '$/release'
        self.rest_items_search = self.rest_items + 'search?'
        self.rest_items_search_scroll = self.rest_items + 'search/scroll?'

        self.params = {'format': 'json', 'scroll': 'true'}
        self.header = {"accept": "application/json", "Content-Type": "application/json"}
        self.query = {"query": {"term": {"context.objectId": {"value": "ctx_924547", "boost": 1.0}}},
                      "size": "50", "from": "0"}

    def count_items(self, query=None):
        self.records = []
        url = self.rest_items_search  # + params
        params = self.params
        headers = self.header
        if query:
            response = utils.post_request(url, params, headers, query)
            if 'numberOfRecords' in response:
                return response['numberOfRecords']
            else:
                print('something went wrong while data retrieval!')
                return 0
        else:
            print("please pass query!")
            return 0

    def search_items(self, query=None):
        self.records = []
        url = self.rest_items_search  # + params
        params = self.params
        headers = self.header
        if query:
            data = query
        else:
            data = self.query
        response = utils.post_request(url, params, headers, data)
        if 'scrollId' in response:
            scroll_id = response['scrollId']
            if 'records' in response:
                self.records += response['records']
                self.scroll_items(scroll_id, 1)
            else:
                return response
        else:
            return response

    # func to repeatedly request items
    def scroll_items(self, scroll_id, counter):
        print("scrolling for the "+str(counter)+". time...")
        headers = self.header
        params = self.format
        params['scrollId'] = scroll_id
        url = self.rest_items_search_scroll
        response = utils.get_request(url, params, headers)
        if 'records' in response:
            self.records += response['records']
            self.scroll_items(response['scrollId'], counter+1)

    def update_item(self, item_id, data):
        if self.auth:
            headers = self.content
            headers['Authorization'] = self.token
            url = self.rest_items + item_id
            return utils.put_request(url, headers, data)
        else:
            print("you need to be authorized to update items!")
            return None

    def release_item(self, item_id, data, comment):
        if self.auth:
            headers = self.content
            headers['Authorization'] = self.token
            url = self.rest_items_release.replace('$', item_id)
            params = {'comment': comment, 'lastModificationDate': data['lastModificationDate']}
            passphrase = self.secret['user-pass'].split(":")[1]
            params['password'] = passphrase
            return utils.put_request(url, headers, params)
        else:
            print("you need to be authorized to release items!")
            return None

    def revise_item(self):
        pass

    def submit_item(self):
        pass

    def withdraw_item(self):
        pass


class OrgUnitRestController(RestController):

    def __init__(self):

        super().__init__()

        # organisational units endpoint
        self.rest_org_units = self.rest + 'ous/'
        self.rest_org_units_search = self.rest_org_units + 'search/'
        self.rest_org_units_children = self.rest_org_units + '$/children'
        self.rest_org_units_toplevel = self.rest_org_units + 'toplevel'

    def get_all(self):
        res = utils.get_request(self.rest_org_units)
        size = {"size": res['numberOfRecords']}
        return utils.get_request(self.rest_org_units + "?", size)

    def get(self, ou_id='ou_907574'):
        url = self.rest_org_units + ou_id
        return utils.get_request(url)

    def top_level(self):
        return utils.get_request(self.rest_org_units_toplevel)

    def child_organizations(self, ou_id='ou_907574'):
        url = self.rest_org_units_children.replace('$', ou_id)
        return utils.get_request(url)


class ContextRestController(RestController):

    def __init__(self):

        super().__init__()

        # context endpoint
        self.rest_contexts = self.rest + 'contexts/'

    def get(self, ctx_id='ctx_924547'):
        url = self.rest_contexts + ctx_id
        return utils.get_request(url)

    def get_all(self):
        res = utils.get_request(self.rest_contexts)
        size = {"size": res['numberOfRecords']}
        return utils.get_request(self.rest_contexts + "?", size)


class FeedRestController(RestController):

    def __init__(self):

        super().__init__()

        self.rest_feed = self.rest + 'feed/'
        self.rest_feed_oa = self.rest_feed + 'oa'
        self.rest_feed_organization = self.rest_feed + 'organization/$'
        self.rest_feed_recent = self.rest_feed + 'recent'
        self.rest_feed_search = self.rest_feed + 'search'

    def get_recent_oa(self):
        pass

    def get_recent_releases_for_ou(self):
        pass

    def get_recent_releases(self):
        pass

    def get_search_as_feed(self):
        pass


class ConeController(BaseController):

    def __init__(self):

        super().__init__()

        # cone interface
        self.cone = self.url + 'cone/'


class PersonConeController(ConeController):

    def __init__(self):

        super().__init__()

        # persons endpoint
        self.cone_persons = self.cone + 'persons/'
        self.cone_persons_all = self.cone_persons + 'all?'
        self.cone_persons_query = self.cone_persons + 'query?'
        self.cone_persons_resource = self.cone_persons + 'resource/'

        self.idx = []
        self.meta = {}
        self.names = {}

        self.dc_idx = 'http_purl_org_dc_elements_1_1_identifier'
        self.dc_title = 'http_purl_org_dc_elements_1_1_title'
        self.dc_alternative = 'http_purl_org_dc_terms_alternative'
        self.escidoc_pos = 'http_purl_org_escidoc_metadata_terms_0_1_position'

        self.nodes = [["Id", "Label"]]
        self.edges = [["Source", "Target"]]
        self.no_edge = []

    def get_entities(self):
        return utils.get_request(self.cone_persons_all, self.format)

    def get_entity(self, person_id='persons32341'):
        url = self.cone_persons_resource + person_id + '?'
        return utils.get_request(url, self.format)

    def init(self):
        pers = self.get_entities()
        for p in pers:
            idx = p['id'].split("/")[-1:][0]
            if idx in self.names:
                self.names[idx].append(p['value'])
            else:
                self.names[idx] = [p['value']]
        self.idx = list(self.names.keys())
        self.idx.sort()

    def full_init(self):
        if not self.meta and not self.names:
            self.init()
        for idx in tqdm(self.idx):
            pers_details = self.get_entity(idx)
            self.meta[idx] = pers_details

    def ous_graph(self):
        #
        # check if data needs to be requested
        #
        if not self.idx and not self.meta:
            self.full_init()
        #
        # iterate over identifiers from persons
        #
        for idx in self.idx:
            pers_details = self.meta[idx]
            #
            # extract name (label) of person (node)
            #
            if self.dc_title in pers_details:
                pers_name = pers_details[self.dc_title]
            elif self.dc_title not in pers_details and self.dc_alternative in pers_details:
                # print(idx, "has just an alternative name!")
                pers_name = pers_details[self.dc_alternative]
            else:
                print("no name found for", idx)
                pers_name = 'NONE'
            self.nodes.append([idx, pers_name])
            #
            # extract affiliation to organizational units (edge)
            #
            if self.escidoc_pos in pers_details:
                affiliation = pers_details[self.escidoc_pos]
                #
                # person has one affiliation
                #
                if type(affiliation) == dict:
                    if self.dc_idx in affiliation:
                        pers_ou = affiliation[self.dc_idx]
                        self.edges.append([idx, pers_ou])
                    else:
                        self.no_edge.append(idx)
                #
                # ... has more than one affilation
                #
                elif type(affiliation) == list:
                    found = False
                    for affil in affiliation:
                        if self.dc_idx in affil:
                            pers_ou = affil[self.dc_idx]
                            self.edges.append([idx, pers_ou])
                            found = True
                    if not found:
                        self.no_edge.append(idx)
                #
                # unhandled data type of affilation
                #
                else:
                    self.no_edge.append(idx)
            #
            # person has no affilation at all
            #
            else:
                self.no_edge.append(idx)

    def search_entity(self, cone_id):
        pass


class JournalConeController(ConeController):

    def __init__(self):

        super().__init__()

        # journals endpoint
        self.cone_journals = self.cone + 'journals/'
        self.cone_journals_all = self.cone_journals + 'all?'
        self.cone_journals_query = self.cone_journals + 'query?'
        self.cone_journals_resource = self.cone_journals + 'resource/'

    def get_entities(self):
        return utils.get_request(self.cone_journals_all, self.format)

    def get_entity(self, journal_id='954928372938'):
        url = self.cone_journals_resource + journal_id + '?'
        return utils.get_request(url, self.format)

    def search_entity(self, cone_id):
        pass


class LanguageConeController(ConeController):

    def __init__(self):

        super().__init__()

        # iso 639-3 endpoint
        self.cone_languages = self.cone + 'iso639-3/'
        self.cone_languages_all = self.cone_languages + 'all?'
        self.cone_language_query = self.cone_languages + 'query?'
        self.cone_languages_resource = self.cone_languages + 'resource/'

        self.idx = []
        self.meta = {}
        self.full_meta = {}

    def get_entities(self):
        return utils.get_request(self.cone_languages_all, self.format)

    def get_entity(self, language_id='deu'):
        url = self.cone_languages_resource + language_id + '?'
        return utils.get_request(url, self.format)

    def init(self):
        langs = self.get_entities()
        for lang in langs:
            idx = lang['id'].split("/")[-1]
            self.meta[idx] = lang['value']
        self.idx = list(self.meta.keys())
        self.idx.sort()

    def full_init(self):
        if not self.idx and not self.meta:
            self.init()
        for idx in tqdm(self.idx):
            lang_details = self.get_entity(idx)
            self.full_meta[idx] = lang_details

    def search_entity(self, cone_id):
        pass
